@@ -519,41 +519,6 @@ authentic_erase_binary(struct sc_card *card, unsigned int offs, size_t count, un
 }
 
 
-#if 0
-static int
-authentic_resize_file(struct sc_card *card, unsigned file_id, unsigned new_size)
-{
-	struct sc_context *ctx = card->ctx;
-	struct sc_apdu apdu;
-	unsigned char data[6] = {
-		0x62, 0x04, 0x80, 0x02, 0xFF, 0xFF
-	};
-	int rv;
-
-	LOG_FUNC_CALLED(ctx);
-	sc_log(ctx, "try to set file size to %i bytes", new_size);
-
-	data[4] = (new_size >> 8) & 0xFF;
-	data[5] = new_size & 0xFF;
-
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0xDB, (file_id >> 8) & 0xFF, file_id & 0xFF);
-	apdu.data = data;
-	apdu.datalen = sizeof(data);
-	apdu.lc = sizeof(data);
-
-	rv = sc_transmit_apdu(card, &apdu);
-	LOG_TEST_RET(ctx, rv, "APDU transmit failed");
-	rv = sc_check_sw(card, apdu.sw1, apdu.sw2);
-	LOG_TEST_RET(ctx, rv, "resize file failed");
-
-	if (card->cache.valid && card->cache.current_ef && card->cache.current_ef->id == file_id)
-		card->cache.current_ef->size = new_size;
-
-	LOG_FUNC_RETURN(ctx, rv);
-}
-#endif
-
-
 static int
 authentic_set_current_files(struct sc_card *card, struct sc_path *path,
 		unsigned char *resp, size_t resplen, struct sc_file **file_out)
@@ -685,7 +650,7 @@ authentic_reduce_path(struct sc_card *card, struct sc_path *path)
 
 	LOG_FUNC_CALLED(ctx);
 
-	if (path->len <= 2 || path->type == SC_PATH_TYPE_DF_NAME || !path)
+	if (!path || path->len <= 2 || path->type == SC_PATH_TYPE_DF_NAME)
 		LOG_FUNC_RETURN(ctx, SC_SUCCESS);
 
 	if (!card->cache.valid || !card->cache.current_df)
@@ -695,7 +660,7 @@ authentic_reduce_path(struct sc_card *card, struct sc_path *path)
 	cur_path = card->cache.current_df->path;
 
 	if (!memcmp(cur_path.value, "\x3F\x00", 2) && memcmp(in_path.value, "\x3F\x00", 2))   {
-		memcpy(in_path.value + 2, in_path.value, in_path.len);
+		memmove(in_path.value + 2, in_path.value, in_path.len);
 		memcpy(in_path.value, "\x3F\x00", 2);
 		in_path.len += 2;
 	}
@@ -707,7 +672,7 @@ authentic_reduce_path(struct sc_card *card, struct sc_path *path)
 			break;
 	}
 
-	memcpy(in_path.value, in_path.value + offs, sizeof(in_path.value) - offs);
+	memmove(in_path.value, in_path.value + offs, sizeof(in_path.value) - offs);
 	in_path.len -= offs;
 	*path = in_path;
 
@@ -784,7 +749,7 @@ authentic_select_file(struct sc_card *card, const struct sc_path *path,
 		rv = authentic_select_mf(card, file_out);
 		LOG_TEST_RET(ctx, rv, "cannot select MF");
 
-		memcpy(&lpath.value[0], &lpath.value[2], lpath.len - 2);
+		memmove(&lpath.value[0], &lpath.value[2], lpath.len - 2);
 		lpath.len -=  2;
 
 		if (!lpath.len)
@@ -1689,9 +1654,7 @@ authentic_pin_reset(struct sc_card *card, struct sc_pin_cmd_data *data, int *tri
 	}
 
 	if (save_current)   {
-		struct sc_file *dummy_file = NULL;
-
-		rv = authentic_select_file(card, &save_current->path, &dummy_file);
+		rv = authentic_select_file(card, &save_current->path, NULL);
 		LOG_TEST_RET(ctx, rv, "Cannot return to saved PATH");
 	}
 	LOG_FUNC_RETURN(ctx, rv);
@@ -2074,9 +2037,6 @@ authentic_set_security_env(struct sc_card *card,
 	default:
 		LOG_FUNC_RETURN(ctx, SC_ERROR_NOT_SUPPORTED);
 	}
-#if 0
-	apdu.flags |= SC_APDU_FLAGS_CAN_WAIT;
-#endif
 
 	rv = sc_transmit_apdu(card, &apdu);
 	LOG_TEST_RET(ctx, rv, "APDU transmit failed");
@@ -2150,14 +2110,16 @@ static int
 authentic_sm_acl_init (struct sc_card *card, struct sm_info *sm_info, int cmd,
 		unsigned char *resp, size_t *resp_len)
 {
-	struct sc_context *ctx = card->ctx;
-	struct sm_type_params_gp *params_gp = &sm_info->session.gp.params;
+	struct sc_context *ctx;
+	struct sm_type_params_gp *params_gp;
 	struct sc_remote_data rdata;
 	int rv;
 
-	sc_log(ctx, "called; command 0x%X\n", cmd);
 	if (!card || !sm_info || !resp || !resp_len)
-		LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_ARGUMENTS);
+		return SC_ERROR_INVALID_ARGUMENTS;
+
+	ctx = card->ctx;
+	params_gp = &sm_info->session.gp.params;
 
 	if (!card->sm_ctx.module.ops.initialize || !card->sm_ctx.module.ops.get_apdus)
 		LOG_FUNC_RETURN(ctx, SC_ERROR_NOT_SUPPORTED);
@@ -2302,22 +2264,22 @@ authentic_sm_get_wrapped_apdu(struct sc_card *card, struct sc_apdu *plain, struc
 
 	LOG_FUNC_CALLED(ctx);
 
-        if (!plain || !sm_apdu)
+	if (!plain || !sm_apdu)
 		LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_ARGUMENTS);
 	sc_log(ctx, "called; CLA:%X, INS:%X, P1:%X, P2:%X, data(%i) %p",
 			plain->cla, plain->ins, plain->p1, plain->p2, plain->datalen, plain->data);
-        *sm_apdu = NULL;
+	*sm_apdu = NULL;
 
 	if ((plain->cla & 0x04)
-		|| (plain->cla==0x00 && plain->ins==0x22)
-		|| (plain->cla==0x00 && plain->ins==0x2A)
-		|| (plain->cla==0x00 && plain->ins==0x84)
-		|| (plain->cla==0x00 && plain->ins==0x88)
-		|| (plain->cla==0x00 && plain->ins==0xA4)
-		|| (plain->cla==0x00 && plain->ins==0xC0)
-		|| (plain->cla==0x00 && plain->ins==0xCA)
-		|| (plain->cla==0x80 && plain->ins==0x50)
-		)   {
+			|| (plain->cla==0x00 && plain->ins==0x22)
+			|| (plain->cla==0x00 && plain->ins==0x2A)
+			|| (plain->cla==0x00 && plain->ins==0x84)
+			|| (plain->cla==0x00 && plain->ins==0x88)
+			|| (plain->cla==0x00 && plain->ins==0xA4)
+			|| (plain->cla==0x00 && plain->ins==0xC0)
+			|| (plain->cla==0x00 && plain->ins==0xCA)
+			|| (plain->cla==0x80 && plain->ins==0x50)
+	   )   {
 		sc_log(ctx, "SM wrap is not applied for this APDU");
 		LOG_FUNC_RETURN(ctx, SC_ERROR_SM_NOT_APPLIED);
 	}
@@ -2328,25 +2290,31 @@ authentic_sm_get_wrapped_apdu(struct sc_card *card, struct sc_apdu *plain, struc
 	if (!card->sm_ctx.module.ops.get_apdus)
 		LOG_FUNC_RETURN(ctx, SC_ERROR_NOT_SUPPORTED);
 
-        apdu = calloc(1, sizeof(struct sc_apdu));
-        if (!apdu)
+	apdu = calloc(1, sizeof(struct sc_apdu));
+	if (!apdu)
 		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
 	memcpy((void *)apdu, (void *)plain, sizeof(struct sc_apdu));
 
-        apdu->data = calloc (1, plain->datalen + 24);
-        if (!apdu->data)
+	apdu->data = calloc (1, plain->datalen + 24);
+	if (!apdu->data)
 		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
 	if (plain->data && plain->datalen)
 		memcpy((unsigned char *) apdu->data, plain->data, plain->datalen);
 
-        apdu->resp = calloc (1, plain->resplen + 32);
-        if (!apdu->resp)
+	apdu->resp = calloc (1, plain->resplen + 32);
+	if (!apdu->resp) {
+		free(apdu);
 		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
+	}
 
 	card->sm_ctx.info.cmd = SM_CMD_APDU_TRANSMIT;
 	card->sm_ctx.info.cmd_data = (void *)apdu;
 
 	rv = card->sm_ctx.module.ops.get_apdus(ctx, &card->sm_ctx.info, NULL, 0, NULL);
+	if (rv < 0) {
+		free(apdu->resp);
+		free(apdu);
+	}
 	LOG_TEST_RET(ctx, rv, "SM: GET_APDUS failed");
 
 	*sm_apdu = apdu;
