@@ -17,7 +17,9 @@
 /* Initially written by David Mattes (david.mattes@boeing.com) */
 /* Portuguese eID card support by Joao Poupino (joao.poupino@ist.utl.pt) */
 
+#if HAVE_CONFIG_H
 #include "config.h"
+#endif
 
 #include <stdlib.h>
 #include <string.h>
@@ -369,7 +371,7 @@ static u8 gemsafe_flags2algref(struct sc_card *card, const struct sc_security_en
 	} else if (env->operation == SC_SEC_OPERATION_DECIPHER) {
 		if (env->algorithm_flags & SC_ALGORITHM_RSA_PAD_PKCS1)
 			ret = (card->type == SC_CARD_TYPE_GEMSAFEV1_PTEID ||
-			       card->type == SC_CARD_TYPE_GEMSAFEV1_PTEID) ? 0x02 : 0x12;
+			       card->type == SC_CARD_TYPE_GEMSAFEV1_SEEID) ? 0x02 : 0x12;
 	}
 
 	return ret;
