@@ -25,7 +25,9 @@
  * http://www.g10code.de/docs/openpgp-card-2.0.pdf
  */
 
+#if HAVE_CONFIG_H
 #include "config.h"
+#endif
 
 #include <stdlib.h>
 #include <string.h>
@@ -305,7 +307,7 @@ pgp_init(sc_card_t *card)
 	sc_file_t	*file = NULL;
 	struct do_info	*info;
 	int		r;
-	struct blob 	*child = NULL;
+	struct blob	*child = NULL;
 
 	priv = calloc (1, sizeof *priv);
 	if (!priv)
@@ -330,8 +332,14 @@ pgp_init(sc_card_t *card)
 		return r;
 	}
 
+	/* defensive programming check */
+	if (!file)   {
+		pgp_finish(card);
+		return SC_ERROR_OBJECT_NOT_FOUND;
+	}
+
 	/* read information from AID */
-	if (file && file->namelen == 16) {
+	if (file->namelen == 16) {
 		/* OpenPGP card spec 1.1 & 2.0, section 4.2.1 & 4.1.2.1 */
 		priv->bcd_version = bebytes2ushort(file->name + 6);
 		/* kludge: get card's serial number from manufacturer ID + serial number */
@@ -883,7 +891,7 @@ static unsigned int pgp_strip_path(sc_card_t *card, const sc_path_t *path)
 {
 	unsigned int start_point = 0;
 	/* start_point will move through the path string */
-	if (path->value == NULL || path->len == 0)
+	if (path->len == 0)
 		return 0;
 
 	/* Ignore 3F00 (MF) at the beginning */
@@ -1873,7 +1881,10 @@ pgp_build_tlv(sc_context_t *ctx, unsigned int tag, u8 *data, size_t len, u8 **ou
 		highest_order++;
 	}
 	highest_order--;
-	cla = tag >> 8*highest_order;
+	if (highest_order >= 4)
+	   cla = 0x00;
+	else	
+		cla = tag >> 8*highest_order;
 	/* Restore class bits */
 	*out[0] |= cla;
 	return SC_SUCCESS;
@@ -2163,7 +2174,9 @@ out:
 /* ABI: card ctl: perform special card-specific operations */
 static int pgp_card_ctl(sc_card_t *card, unsigned long cmd, void *ptr)
 {
+#ifdef ENABLE_OPENSSL
 	int r;
+#endif /* ENABLE_OPENSSL */
 
 	LOG_FUNC_CALLED(card->ctx);
 
