@@ -63,15 +63,17 @@ static int
 cosm_write_tokeninfo (struct sc_pkcs15_card *p15card, struct sc_profile *profile,
 		char *label, unsigned flags)
 {
-	struct sc_context *ctx = p15card->card->ctx;
+	struct sc_context *ctx;
 	struct sc_file *file = NULL;
 	int rv;
 	size_t sz;
 	char *buffer = NULL;
 
-	if (!p15card || !profile)
+	if (!p15card || !p15card->card || !profile)
 		return SC_ERROR_INVALID_ARGUMENTS;
 
+	ctx = p15card->card->ctx;
+
 	SC_FUNC_CALLED(ctx, SC_LOG_DEBUG_VERBOSE);
 	sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "cosm_write_tokeninfo() label '%s'; flags 0x%X", label, flags);
 	if (sc_profile_get_file(profile, COSM_TITLE"-token-info", &file))
@@ -228,24 +230,11 @@ cosm_create_dir(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 		COSM_TITLE"-container-list",
 		COSM_TITLE"-public-list",
 		COSM_TITLE"-private-list",
-#if 0
-		"PKCS15-AppDF",
-		"PKCS15-ODF",
-		"PKCS15-AODF",
-		"PKCS15-PrKDF",
-		"PKCS15-PuKDF",
-		"PKCS15-CDF",
-		"PKCS15-DODF",
-#endif
 		NULL
 	};
 
 	SC_FUNC_CALLED(ctx, SC_LOG_DEBUG_VERBOSE);
 
-#if 0
-	rv = sc_pkcs15init_create_file(profile, p15card, df);
-	SC_TEST_RET(ctx, SC_LOG_DEBUG_NORMAL, rv, "Failed to create DIR DF");
-#endif
 	/* Oberthur AWP file system is expected.*/
 	/* Create private objects DF */
 	for (ii = 0; create_dfs[ii]; ii++)   {
@@ -277,7 +266,6 @@ cosm_create_reference_data(struct sc_profile *profile, struct sc_pkcs15_card *p1
 	struct sc_card *card = p15card->card;
 	struct sc_pkcs15_auth_info profile_auth_pin, profile_auth_puk;
 	struct sc_cardctl_oberthur_createpin_info args;
-	unsigned char *puk_buff = NULL;
 	int rv;
 	unsigned char oberthur_puk[16] = {
 		0x6F, 0x47, 0xD9, 0x88, 0x4B, 0x6F, 0x9D, 0xC5,
@@ -334,9 +322,6 @@ cosm_create_reference_data(struct sc_profile *profile, struct sc_pkcs15_card *p1
 			sc_file_free(file);
 	}
 
-	if (puk_buff)
-		free(puk_buff);
-
 	SC_FUNC_RETURN(ctx, SC_LOG_DEBUG_NORMAL, rv);
 }
 
