@@ -62,6 +62,8 @@ static int	verbose = 0;
 #define MAX_KEY			1024
 #define MAX_WRAPPED_KEY	(MAX_CERT + MAX_PRKD + MAX_KEY)
 
+#define SEED_LENGTH 16
+
 enum {
 	OPT_SO_PIN = 0x100,
 	OPT_PIN,
@@ -139,14 +141,15 @@ static sc_card_t *card = NULL;
  * @param s Secret to share
  * @param n Maximum number of shares
  * @param rngSeed Seed value for CPRNG
+ * @param rngSeedLength Lenght of Seed value for CPRNG
  *
  */
-static void generatePrime(BIGNUM *prime, const BIGNUM *s, const unsigned int n, unsigned char *rngSeed)
+static void generatePrime(BIGNUM *prime, const BIGNUM *s, const unsigned int n, unsigned char *rngSeed, const unsigned int rngSeedLength)
 {
 	int bits = 0;
 
 	// Seed the RNG
-	RAND_seed(rngSeed, sizeof(rngSeed));
+	RAND_seed(rngSeed, rngSeedLength);
 
 	// Determine minimum number of bits for prime >= max(2^r, n + 1)
 	bits = BN_num_bits_word(n + 1) > BN_num_bits(s) ? (BN_num_bits_word(n + 1)) : (BN_num_bits(s));
@@ -342,6 +345,7 @@ static int reconstructSecret(secret_share_t *shares, unsigned char t, const BIGN
 		 * multiplication
 		 */
 		if (BN_mod_inverse(&denominator, &denominator, &prime, ctx) == NULL ) {
+			free(bValue);
 			return -1;
 		}
 
@@ -417,14 +421,18 @@ static int cleanUpShares(secret_share_t *shares, unsigned char n)
 
 void clearScreen()
 {
-	if (system( "clear" )) system( "cls" );
+	if (system( "clear" )) {
+		if (system( "cls" )) {
+			fprintf(stderr, "Clearing the screen failed\n");
+		}
+	}
 }
 
 
 
 void waitForEnterKeyPressed()
 {
-	char c;
+	int c;
 
 	fflush(stdout);
 	while ((c = getchar()) != '\n' && c != EOF) {
@@ -493,7 +501,7 @@ static void print_info(sc_card_t *card, sc_file_t *file)
 
 
 
-static void initialize(sc_card_t *card, const char *so_pin, const char *user_pin, int retry_counter, int dkek_shares, const char *label)
+static int initialize(sc_card_t *card, const char *so_pin, const char *user_pin, int retry_counter, int dkek_shares, const char *label)
 {
 	sc_cardctl_sc_hsm_init_param_t param;
 	size_t len;
@@ -512,12 +520,12 @@ static void initialize(sc_card_t *card, const char *so_pin, const char *user_pin
 	r = sc_hex_to_bin(_so_pin, param.init_code, &len);
 	if (r < 0) {
 		fprintf(stderr, "Error decoding initialization code (%s)\n", sc_strerror(r));
-		return;
+		return -1;
 	}
 
 	if (len != 8) {
 		fprintf(stderr, "SO-PIN must be a hexadecimal string of 16 characters\n");
-		return;
+		return -1;
 	}
 
 	if (user_pin == NULL) {
@@ -532,27 +540,27 @@ static void initialize(sc_card_t *card, const char *so_pin, const char *user_pin
 
 	if (param.user_pin_len < 6) {
 		fprintf(stderr, "PIN must be at least 6 characters long\n");
-		return;
+		return -1;
 	}
 
 	if (param.user_pin_len > 16) {
 		fprintf(stderr, "PIN must not be longer than 16 characters\n");
-		return;
+		return -1;
 	}
 
 	if ((param.user_pin_len == 6) && (retry_counter > 3)) {
 		fprintf(stderr, "Retry counter must not exceed 3 for a 6 digit PIN. Use a longer PIN for a higher retry counter.\n");
-		return;
+		return -1;
 	}
 
 	if ((param.user_pin_len == 7) && (retry_counter > 5)) {
 		fprintf(stderr, "Retry counter must not exceed 5 for a 7 digit PIN. Use a longer PIN for a higher retry counter.\n");
-		return;
+		return -1;
 	}
 
 	if (retry_counter > 10) {
 		fprintf(stderr, "Retry counter must not exceed 10\n");
-		return;
+		return -1;
 	}
 
 	param.user_pin = (u8 *)_user_pin;
@@ -569,6 +577,8 @@ static void initialize(sc_card_t *card, const char *so_pin, const char *user_pin
 	if (r < 0) {
 		fprintf(stderr, "sc_card_ctl(*, SC_CARDCTL_SC_HSM_INITIALIZE, *) failed with %s\n", sc_strerror(r));
 	}
+
+	return 0;
 }
 
 
@@ -586,6 +596,11 @@ static int recreate_password_from_shares(char **pwd, int *pwdlen, int num_of_pas
 	secret_share_t *shares = NULL;
 	secret_share_t *sp;
 
+	if (num_of_password_shares < 2) {
+		fprintf(stderr, "--pwd-shares-total must 2 or larger\n");
+		return -1;
+	}
+
 	/*
 	 * Initialize prime and secret
 	 */
@@ -600,7 +615,10 @@ static int recreate_password_from_shares(char **pwd, int *pwdlen, int num_of_pas
 	printf("\nPlease remember to present the share id as well as the share value.");
 	printf("\n\nPlease enter prime: ");
 	memset(inbuf, 0, sizeof(inbuf));
-	fgets(inbuf, sizeof(inbuf), stdin);
+	if (fgets(inbuf, sizeof(inbuf), stdin) == NULL) {
+		fprintf(stderr, "Input aborted\n");
+		return -1;
+	}
 	binlen = 64;
 	sc_hex_to_bin(inbuf, bin, &binlen);
 	BN_bin2bn(bin, binlen, &prime);
@@ -621,13 +639,19 @@ static int recreate_password_from_shares(char **pwd, int *pwdlen, int num_of_pas
 
 		printf("Please enter share ID: ");
 		memset(inbuf, 0, sizeof(inbuf));
-		fgets(inbuf, sizeof(inbuf), stdin);
+		if (fgets(inbuf, sizeof(inbuf), stdin) == NULL) {
+			fprintf(stderr, "Input aborted\n");
+			return -1;
+		}
 		p = &(sp->x);
 		BN_hex2bn(&p, inbuf);
 
 		printf("Please enter share value: ");
 		memset(inbuf, 0, sizeof(inbuf));
-		fgets(inbuf, sizeof(inbuf), stdin);
+		if (fgets(inbuf, sizeof(inbuf), stdin) == NULL) {
+			fprintf(stderr, "Input aborted\n");
+			return -1;
+		}
 		binlen = 64;
 		sc_hex_to_bin(inbuf, bin, &binlen);
 		BN_bin2bn(bin, binlen, &(sp->y));
@@ -641,6 +665,7 @@ static int recreate_password_from_shares(char **pwd, int *pwdlen, int num_of_pas
 
 	if (r < 0) {
 		printf("\nError during reconstruction of secret. Wrong shares?\n");
+		cleanUpShares(shares, num_of_password_shares);
 		return r;
 	}
 
@@ -662,7 +687,7 @@ static int recreate_password_from_shares(char **pwd, int *pwdlen, int num_of_pas
 
 
 
-static void import_dkek_share(sc_card_t *card, const char *inf, int iter, char *password, int num_of_password_shares)
+static int import_dkek_share(sc_card_t *card, const char *inf, int iter, const char *password, int num_of_password_shares)
 {
 	sc_cardctl_sc_hsm_dkek_t dkekinfo;
 	EVP_CIPHER_CTX ctx;
@@ -671,23 +696,29 @@ static void import_dkek_share(sc_card_t *card, const char *inf, int iter, char *
 	char *pwd = NULL;
 	int r, outlen, pwdlen;
 
+	if (inf == NULL) {
+		fprintf(stderr, "No file name specified for DKEK share\n");
+		return -1;
+	}
+
 	in = fopen(inf, "rb");
 
 	if (in == NULL) {
 		perror(inf);
-		return;
+		return -1;
 	}
 
 	if (fread(filebuff, 1, sizeof(filebuff), in) != sizeof(filebuff)) {
 		perror(inf);
-		return;
+		fclose(in);
+		return -1;
 	}
 
 	fclose(in);
 
 	if (memcmp(filebuff, magic, sizeof(magic) - 1)) {
-		printf("File %s is not a DKEK share\n", inf);
-		return;
+		fprintf(stderr, "File %s is not a DKEK share\n", inf);
+		return -1;
 	}
 
 	if (password == NULL) {
@@ -700,12 +731,12 @@ static void import_dkek_share(sc_card_t *card, const char *inf, int iter, char *
 		} else {
 			r = recreate_password_from_shares(&pwd, &pwdlen, num_of_password_shares);
 			if (r < 0) {
-				return;
+				return -1;
 			}
 		}
 
 	} else {
-		pwd = password;
+		pwd = (char *) password;
 		pwdlen = strlen(password);
 	}
 
@@ -720,13 +751,13 @@ static void import_dkek_share(sc_card_t *card, const char *inf, int iter, char *
 	EVP_CIPHER_CTX_init(&ctx);
 	EVP_DecryptInit_ex(&ctx, EVP_aes_256_cbc(), NULL, key, iv);
 	if (!EVP_DecryptUpdate(&ctx, outbuff, &outlen, filebuff + 16, sizeof(filebuff) - 16)) {
-		printf("Error decrypting DKEK share. Password correct ?\n");
-		return;
+		fprintf(stderr, "Error decrypting DKEK share. Password correct ?\n");
+		return -1;
 	}
 
 	if (!EVP_DecryptFinal_ex(&ctx, outbuff + outlen, &r)) {
-		printf("Error decrypting DKEK share. Password correct ?\n");
-		return;
+		fprintf(stderr, "Error decrypting DKEK share. Password correct ?\n");
+		return -1;
 	}
 
 	memset(&dkekinfo, 0, sizeof(dkekinfo));
@@ -741,16 +772,17 @@ static void import_dkek_share(sc_card_t *card, const char *inf, int iter, char *
 	EVP_CIPHER_CTX_cleanup(&ctx);
 
 	if (r == SC_ERROR_INS_NOT_SUPPORTED) {			// Not supported or not initialized for key shares
-		printf("Not supported by card or card not initialized for key share usage\n");
-		return;
+		fprintf(stderr, "Not supported by card or card not initialized for key share usage\n");
+		return -1;
 	}
 
 	if (r < 0) {
 		fprintf(stderr, "sc_card_ctl(*, SC_CARDCTL_SC_HSM_IMPORT_DKEK_SHARE, *) failed with %s\n", sc_strerror(r));
-		return;
+		return -1;
 	}
 	printf("DKEK share imported\n");
 	print_dkek_info(&dkekinfo);
+	return 0;
 }
 
 
@@ -806,6 +838,26 @@ static int generate_pwd_shares(sc_card_t *card, char **pwd, int *pwdlen, int pas
 
 	u8 rngseed[16];
 
+	if ((password_shares_threshold == -1) || (password_shares_total == -1)) {
+		fprintf(stderr, "Must specify both, --pwd-shares-total and --pwd-shares-threshold\n");
+		return -1;
+	}
+
+	if (password_shares_total < 3) {
+		fprintf(stderr, "--pwd-shares-total must be 3 or larger\n");
+		return -1;
+	}
+
+	if (password_shares_threshold < 2) {
+		fprintf(stderr, "--pwd-shares-threshold must 2 or larger\n");
+		return -1;
+	}
+
+	if (password_shares_threshold > password_shares_total) {
+		fprintf(stderr, "--pwd-shares-threshold must be smaller or equal to --pwd-shares-total\n");
+		return -1;
+	}
+
 	printf(	"\nThe DKEK will be enciphered using a randomly generated 64 bit password.\n");
 	printf(	"This password is split using a (%i-of-%i) threshold scheme.\n\n", password_shares_threshold, password_shares_total);
 
@@ -843,7 +895,7 @@ static int generate_pwd_shares(sc_card_t *card, char **pwd, int *pwdlen, int pas
 	/*
 	 * Generate seed and calculate a prime depending on the size of the secret
 	 */
-	r = sc_get_challenge(card, rngseed, 16);
+	r = sc_get_challenge(card, rngseed, SEED_LENGTH);
 	if (r < 0) {
 		printf("Error generating random seed failed with %s", sc_strerror(r));
 		OPENSSL_cleanse(*pwd, *pwdlen);
@@ -851,7 +903,7 @@ static int generate_pwd_shares(sc_card_t *card, char **pwd, int *pwdlen, int pas
 		return r;
 	}
 
-	generatePrime(&prime, &secret, password_shares_total, rngseed);
+	generatePrime(&prime, &secret, password_shares_total, rngseed, SEED_LENGTH);
 
 	// Allocate data buffer for the generated shares
 	shares = malloc(password_shares_total * sizeof(secret_share_t));
@@ -896,7 +948,7 @@ static int generate_pwd_shares(sc_card_t *card, char **pwd, int *pwdlen, int pas
 
 
 
-static void create_dkek_share(sc_card_t *card, const char *outf, int iter, char *password, int password_shares_threshold, int password_shares_total)
+static int create_dkek_share(sc_card_t *card, const char *outf, int iter, const char *password, int password_shares_threshold, int password_shares_total)
 {
 	EVP_CIPHER_CTX ctx;
 	FILE *out = NULL;
@@ -905,30 +957,34 @@ static void create_dkek_share(sc_card_t *card, const char *outf, int iter, char
 	char *pwd = NULL;
 	int r = 0, outlen, pwdlen = 0;
 
-	if (password == NULL) {
+	if (outf == NULL) {
+		fprintf(stderr, "No file name specified for DKEK share\n");
+		return -1;
+	}
 
-		if (password_shares_threshold == -1) {
+	if (password == NULL) {
+		if ((password_shares_threshold == -1) && (password_shares_total == -1)) {
 			ask_for_password(&pwd, &pwdlen);
 		} else { // create password using threshold scheme
 			r = generate_pwd_shares(card, &pwd, &pwdlen, password_shares_threshold, password_shares_total);
 		}
 
 	} else {
-		pwd = password;
+		pwd = (char *) password;
 		pwdlen = strlen(password);
 	}
 
 	if (r < 0) {
-		printf("Creating DKEK share failed");
-		return;
+		fprintf(stderr, "Creating DKEK share failed\n");
+		return -1;
 	}
 
 	memcpy(filebuff, magic, sizeof(magic) - 1);
 
 	r = sc_get_challenge(card, filebuff + 8, 8);
 	if (r < 0) {
-		printf("Error generating random number failed with %s", sc_strerror(r));
-		return;
+		fprintf(stderr, "Error generating random number failed with %s\n", sc_strerror(r));
+		return -1;
 	}
 
 	printf("Enciphering DKEK share, please wait...\n");
@@ -941,32 +997,33 @@ static void create_dkek_share(sc_card_t *card, const char *outf, int iter, char
 
 	r = sc_get_challenge(card, dkek_share, sizeof(dkek_share));
 	if (r < 0) {
-		printf("Error generating random number failed with %s", sc_strerror(r));
-		return;
+		fprintf(stderr, "Error generating random number failed with %s\n", sc_strerror(r));
+		return -1;
 	}
 
 	EVP_CIPHER_CTX_init(&ctx);
 	EVP_EncryptInit_ex(&ctx, EVP_aes_256_cbc(), NULL, key, iv);
 	if (!EVP_EncryptUpdate(&ctx, filebuff + 16, &outlen, dkek_share, sizeof(dkek_share))) {
-		printf("Error encrypting DKEK share\n");
-		return;
+		fprintf(stderr, "Error encrypting DKEK share\n");
+		return -1;
 	}
 
 	if (!EVP_EncryptFinal_ex(&ctx, filebuff + 16 + outlen, &r)) {
-		printf("Error encrypting DKEK share\n");
-		return;
+		fprintf(stderr, "Error encrypting DKEK share\n");
+		return -1;
 	}
 
 	out = fopen(outf, "wb");
 
 	if (out == NULL) {
 		perror(outf);
-		return;
+		return -1;
 	}
 
 	if (fwrite(filebuff, 1, sizeof(filebuff), out) != sizeof(filebuff)) {
 		perror(outf);
-		return;
+		fclose(out);
+		return -1;
 	}
 
 	fclose(out);
@@ -975,6 +1032,7 @@ static void create_dkek_share(sc_card_t *card, const char *outf, int iter, char
 	EVP_CIPHER_CTX_cleanup(&ctx);
 
 	printf("DKEK share created and saved to %s\n", outf);
+	return 0;
 }
 
 
@@ -1038,7 +1096,7 @@ static int wrap_with_tag(u8 tag, u8 *indata, size_t inlen, u8 **outdata, size_t
 
 
 
-static void wrap_key(sc_card_t *card, u8 keyid, const char *outf, const char *pin)
+static int wrap_key(sc_card_t *card, int keyid, const char *outf, const char *pin)
 {
 	sc_cardctl_sc_hsm_wrapped_key_t wrapped_key;
 	struct sc_pin_cmd_data data;
@@ -1055,6 +1113,16 @@ static void wrap_key(sc_card_t *card, u8 keyid, const char *outf, const char *pi
 	size_t key_len;
 	int r, ef_prkd_len, ef_cert_len;
 
+	if ((keyid < 1) || (keyid > 255)) {
+		fprintf(stderr, "Invalid key reference (must be 0 < keyid <= 255)\n");
+		return -1;
+	}
+
+	if (outf == NULL) {
+		fprintf(stderr, "No file name specified for wrapped key\n");
+		return -1;
+	}
+
 	if (pin == NULL) {
 		printf("Enter User PIN : ");
 		util_getpass(&lpin, NULL, stdin);
@@ -1074,7 +1142,7 @@ static void wrap_key(sc_card_t *card, u8 keyid, const char *outf, const char *pi
 
 	if (r < 0) {
 		fprintf(stderr, "PIN verification failed with %s\n", sc_strerror(r));
-		return;
+		return -1;
 	}
 
 	if (pin == NULL) {
@@ -1088,17 +1156,18 @@ static void wrap_key(sc_card_t *card, u8 keyid, const char *outf, const char *pi
 	r = sc_card_ctl(card, SC_CARDCTL_SC_HSM_WRAP_KEY, (void *)&wrapped_key);
 
 	if (r == SC_ERROR_INS_NOT_SUPPORTED) {			// Not supported or not initialized for key shares
-		return;
+		fprintf(stderr, "Card not initialized for key wrap\n");
+		return -1;
 	}
 
 	if (r < 0) {
 		fprintf(stderr, "sc_card_ctl(*, SC_CARDCTL_SC_HSM_WRAP_KEY, *) failed with %s\n", sc_strerror(r));
-		return;
+		return -1;
 	}
 
 
 	fid[0] = PRKD_PREFIX;
-	fid[1] = keyid;
+	fid[1] = (unsigned char)keyid;
 	ef_prkd_len = 0;
 
 	/* Try to select a related EF containing the PKCS#15 description of the key */
@@ -1117,7 +1186,7 @@ static void wrap_key(sc_card_t *card, u8 keyid, const char *outf, const char *pi
 	}
 
 	fid[0] = EE_CERTIFICATE_PREFIX;
-	fid[1] = keyid;
+	fid[1] = (unsigned char)keyid;
 	ef_cert_len = 0;
 
 	/* Try to select a related EF containing the certificate for the key */
@@ -1144,7 +1213,10 @@ static void wrap_key(sc_card_t *card, u8 keyid, const char *outf, const char *pi
 
 	memcpy(ptr, key, key_len);
 	ptr += key_len;
+
 	free(key);
+	key = NULL;
+	key_len = 0;
 
 	// Add private key description
 	if (ef_prkd_len > 0) {
@@ -1159,7 +1231,6 @@ static void wrap_key(sc_card_t *card, u8 keyid, const char *outf, const char *pi
 	}
 
 	// Encode key, key decription and certificate object in sequence
-	key_len = 0;
 	wrap_with_tag(0x30, keyblob, ptr - keyblob, &key, &key_len);
 
 	out = fopen(outf, "wb");
@@ -1167,17 +1238,19 @@ static void wrap_key(sc_card_t *card, u8 keyid, const char *outf, const char *pi
 	if (out == NULL) {
 		perror(outf);
 		free(key);
-		return;
+		return -1;
 	}
 
 	if (fwrite(key, 1, key_len, out) != key_len) {
 		perror(outf);
 		free(key);
-		return;
+		fclose(out);
+		return -1;
 	}
 
 	free(key);
 	fclose(out);
+	return 0;
 }
 
 
@@ -1221,7 +1294,7 @@ static int update_ef(sc_card_t *card, u8 prefix, u8 id, int erase, const u8 *buf
 
 
 
-static void unwrap_key(sc_card_t *card, u8 keyid, const char *inf, const char *pin, int force)
+static int unwrap_key(sc_card_t *card, int keyid, const char *inf, const char *pin, int force)
 {
 	sc_cardctl_sc_hsm_wrapped_key_t wrapped_key;
 	struct sc_pin_cmd_data data;
@@ -1235,16 +1308,26 @@ static void unwrap_key(sc_card_t *card, u8 keyid, const char *inf, const char *p
 	int r, keybloblen;
 	size_t len, olen, prkd_len, cert_len;
 
+	if ((keyid < 1) || (keyid > 255)) {
+		fprintf(stderr, "Invalid key reference (must be 0 < keyid <= 255)\n");
+		return -1;
+	}
+
+	if (inf == NULL) {
+		fprintf(stderr, "No file name specified for wrapped key\n");
+		return -1;
+	}
+
 	in = fopen(inf, "rb");
 
 	if (in == NULL) {
 		perror(inf);
-		return;
+		return -1;
 	}
 
 	if ((keybloblen = fread(keyblob, 1, sizeof(keyblob), in)) < 0) {
 		perror(inf);
-		return;
+		return -1;
 	}
 
 	fclose(in);
@@ -1254,14 +1337,14 @@ static void unwrap_key(sc_card_t *card, u8 keyid, const char *inf, const char *p
 			((cla & SC_ASN1_TAG_CONSTRUCTED) != SC_ASN1_TAG_CONSTRUCTED) ||
 			((tag != SC_ASN1_TAG_SEQUENCE)) ){
 		fprintf(stderr, "Invalid wrapped key format (Outer sequence).\n");
-		return;
+		return -1;
 	}
 
 	if ((sc_asn1_read_tag(&ptr, len, &cla, &tag, &olen) != SC_SUCCESS) ||
 			(cla & SC_ASN1_TAG_CONSTRUCTED) ||
 			((tag != SC_ASN1_TAG_OCTET_STRING)) ){
 		fprintf(stderr, "Invalid wrapped key format (Key binary).\n");
-		return;
+		return -1;
 	}
 
 	wrapped_key.wrapped_key = (u8 *)ptr;
@@ -1286,7 +1369,7 @@ static void unwrap_key(sc_card_t *card, u8 keyid, const char *inf, const char *p
 
 	if ((prkd_len > 0) && !force) {
 		fid[0] = PRKD_PREFIX;
-		fid[1] = keyid;
+		fid[1] = (unsigned char)keyid;
 
 		/* Try to select a related EF containing the PKCS#15 description of the key */
 		sc_path_set(&path, SC_PATH_TYPE_FILE_ID, fid, sizeof(fid), 0, 0);
@@ -1294,13 +1377,13 @@ static void unwrap_key(sc_card_t *card, u8 keyid, const char *inf, const char *p
 
 		if (r == SC_SUCCESS) {
 			fprintf(stderr, "Found existing private key description in EF with fid %02x%02x. Please remove key first, select unused key reference or use --force.\n", fid[0], fid[1]);
-			return;
+			return -1;
 		}
 	}
 
 	if ((cert_len > 0) && !force) {
 		fid[0] = EE_CERTIFICATE_PREFIX;
-		fid[1] = keyid;
+		fid[1] = (unsigned char)keyid;
 
 		/* Try to select a related EF containing the certificate */
 		sc_path_set(&path, SC_PATH_TYPE_FILE_ID, fid, sizeof(fid), 0, 0);
@@ -1308,7 +1391,7 @@ static void unwrap_key(sc_card_t *card, u8 keyid, const char *inf, const char *p
 
 		if (r == SC_SUCCESS) {
 			fprintf(stderr, "Found existing certificate in EF with fid %02x%02x. Please remove certificate first, select unused key reference or use --force.\n", fid[0], fid[1]);
-			return;
+			return -1;
 		}
 	}
 
@@ -1331,7 +1414,7 @@ static void unwrap_key(sc_card_t *card, u8 keyid, const char *inf, const char *p
 
 	if (r < 0) {
 		fprintf(stderr, "PIN verification failed with %s\n", sc_strerror(r));
-		return;
+		return -1;
 	}
 
 	if (pin == NULL) {
@@ -1351,12 +1434,18 @@ static void unwrap_key(sc_card_t *card, u8 keyid, const char *inf, const char *p
 	r = sc_card_ctl(card, SC_CARDCTL_SC_HSM_UNWRAP_KEY, (void *)&wrapped_key);
 
 	if (r == SC_ERROR_INS_NOT_SUPPORTED) {			// Not supported or not initialized for key shares
-		return;
+		fprintf(stderr, "Card not initialized for key wrap\n");
+		return -1;
+	}
+
+	if (r == SC_ERROR_INCORRECT_PARAMETERS) {			// Not supported or not initialized for key shares
+		fprintf(stderr, "Wrapped key does not match DKEK\n");
+		return -1;
 	}
 
 	if (r < 0) {
 		fprintf(stderr, "sc_card_ctl(*, SC_CARDCTL_SC_HSM_UNWRAP_KEY, *) failed with %s\n", sc_strerror(r));
-		return;
+		return -1;
 	}
 
 	if (prkd_len > 0) {
@@ -1364,7 +1453,7 @@ static void unwrap_key(sc_card_t *card, u8 keyid, const char *inf, const char *p
 
 		if (r < 0) {
 			fprintf(stderr, "Updating private key description failed with %s\n", sc_strerror(r));
-			return;
+			return -1;
 		}
 	}
 
@@ -1373,11 +1462,12 @@ static void unwrap_key(sc_card_t *card, u8 keyid, const char *inf, const char *p
 
 		if (r < 0) {
 			fprintf(stderr, "Updating certificate failed with %s\n", sc_strerror(r));
-			return;
+			return -1;
 		}
 	}
 
 	printf("Key successfully imported\n");
+	return 0;
 }
 
 
@@ -1396,7 +1486,7 @@ int main(int argc, char * const argv[])
 	const char *opt_so_pin = NULL;
 	const char *opt_pin = NULL;
 	const char *opt_filename = NULL;
-	char *opt_password = NULL;
+	const char *opt_password = NULL;
 	int opt_retry_counter = 3;
 	int opt_dkek_shares = -1;
 	int opt_key_reference = -1;
@@ -1441,13 +1531,13 @@ int main(int argc, char * const argv[])
 			action_count++;
 			break;
 		case OPT_PASSWORD:
-			opt_password = optarg;
+			util_get_pin(optarg, &opt_password);
 			break;
 		case OPT_SO_PIN:
-			opt_so_pin = optarg;
+			util_get_pin(optarg, &opt_so_pin);
 			break;
 		case OPT_PIN:
-			opt_pin = optarg;
+			util_get_pin(optarg, &opt_pin);
 			break;
 		case OPT_RETRY:
 			opt_retry_counter = atol(optarg);
@@ -1492,7 +1582,7 @@ int main(int argc, char * const argv[])
 	r = sc_context_create(&ctx, &ctx_param);
 	if (r != SC_SUCCESS) {
 		fprintf(stderr, "Failed to establish context: %s\n", sc_strerror(r));
-		return 1;
+		exit(1);
 	}
 
 	/* Only change if not in opensc.conf */
@@ -1501,9 +1591,9 @@ int main(int argc, char * const argv[])
 		sc_ctx_log_to_file(ctx, "stderr");
 	}
 
-	err = util_connect_card(ctx, &card, opt_reader, opt_wait, verbose);
-	if (err != SC_SUCCESS) {
-		if (err < 0) {
+	r = util_connect_card(ctx, &card, opt_reader, opt_wait, verbose);
+	if (r != SC_SUCCESS) {
+		if (r < 0) {
 			fprintf(stderr, "Failed to connect to card: %s\n", sc_strerror(err));
 		}
 		goto end;
@@ -1514,33 +1604,32 @@ int main(int argc, char * const argv[])
 
 	if (r != SC_SUCCESS) {
 		fprintf(stderr, "Failed to select application: %s\n", sc_strerror(r));
-		goto end;
+		goto fail;
 	}
 
-	if (do_initialize) {
-		initialize(card, opt_so_pin, opt_pin, opt_retry_counter, opt_dkek_shares, opt_label);
-	}
+	if (do_initialize && initialize(card, opt_so_pin, opt_pin, opt_retry_counter, opt_dkek_shares, opt_label))
+		goto fail;
 
-	if (do_create_dkek_share) {
-		create_dkek_share(card, opt_filename, opt_iter, opt_password, opt_password_shares_threshold, opt_password_shares_total);
-	}
+	if (do_create_dkek_share && create_dkek_share(card, opt_filename, opt_iter, opt_password, opt_password_shares_threshold, opt_password_shares_total))
+		goto fail;
 
-	if (do_import_dkek_share) {
-		import_dkek_share(card, opt_filename, opt_iter, opt_password, opt_password_shares_total);
-	}
+	if (do_import_dkek_share && import_dkek_share(card, opt_filename, opt_iter, opt_password, opt_password_shares_total))
+		goto fail;
 
-	if (do_wrap_key) {
-		wrap_key(card, opt_key_reference, opt_filename, opt_pin);
-	}
+	if (do_wrap_key && wrap_key(card, opt_key_reference, opt_filename, opt_pin))
+		goto fail;
 
-	if (do_unwrap_key) {
-		unwrap_key(card, opt_key_reference, opt_filename, opt_pin, opt_force);
-	}
+	if (do_unwrap_key && unwrap_key(card, opt_key_reference, opt_filename, opt_pin, opt_force))
+		goto fail;
 
 	if (action_count == 0) {
 		print_info(card, file);
 	}
 
+	err = 0;
+	goto end;
+fail:
+	err = 1;
 end:
 	if (card) {
 		sc_unlock(card);
